#ifndef _RUN_PART_H
#define _RUN_PART_H

int run_part (char *script_path, const char *name, const char *action);
int run_parts (const char *directory, const char *name, const char *action);

#endif /* _RUN_PART_H */
